/* シャッフルを行う関数とその返り値を格納する構造体 */
import java.util.*;


public class Shuffler {
	private static final int NORMAL = 0, TRASH = 1, ACTION = 2, BUY = 4, BANE = 8;
	public enum TriState { FALSE, TRUE, NO };

	/* TRASH, ACTION, BUY の条件による抽出 */
	private static List<Card> constraints(int gap, boolean trash, boolean action, boolean buy){	// gap = needed - minSum
		if (!(trash || action || buy)) return new LinkedList<Card>();

		int cstrt = trash ? TRASH : action ? ACTION : BUY;
		RandomList<Card> shuffleList = new RandomList<Card>();
		Card tmp;

		/* needed == minSum の場合（minVal > 0 なるセットから抽出せねばならない） */
		if (gap == 0){
			for (Expansion exp : Expansion.allInstances){
				if (exp.minVal <= 0) continue;
				for (Card card : exp.cards) if (card.isAvailable() && (card.attrib & cstrt) != 0) shuffleList.add(card);
			}
			loop1: while (true){
				tmp = shuffleList.randomRemove();	// 空なら例外発生
				try {
					tmp.disabled = true; tmp.expansion.maxVal--; tmp.expansion.minVal--;
					List<Card> ret = constraints(0, false, action && !((tmp.attrib & ACTION) != 0), buy && !((tmp.attrib & BUY) != 0));
					ret.add(tmp);
					return ret;
				} catch (Exception ex){
					tmp.disabled = false; tmp.expansion.maxVal++; tmp.expansion.minVal++;
					continue loop1;
				}
			}

		/* そうでない場合 */
		} else {
			for (Expansion exp : Expansion.allInstances){
				if (!exp.isSelected()) continue;
				for (Card card : exp.cards) if (card.isAvailable() && (card.attrib & cstrt) != 0) shuffleList.add(card);
			}
			loop2: while (true){
				tmp = shuffleList.randomRemove();	// 空なら例外発生
				try {
					tmp.disabled = true; tmp.expansion.maxVal--; tmp.expansion.minVal--;
					List<Card> ret = constraints((tmp.expansion.minVal >= 0) ? gap : (gap - 1),
						false, action && !((tmp.attrib & ACTION) != 0), buy && !((tmp.attrib & BUY) != 0));
					ret.add(tmp);
					return ret;
				} catch (Exception ex){
					tmp.disabled = false; tmp.expansion.maxVal++; tmp.expansion.minVal++;
					continue loop2;
				}
			}
		}
	}

	public static List<Card> shuffle(int needed, boolean trash, boolean action, boolean buy, boolean alchemy) throws ExpectedException {
		List<Card> chosen;
		Card tmp;

		/* 値の初期化 */
		int minSum = 0, maxSum = 0;
		for (Expansion exp : Expansion.allInstances){
			if (exp.isSelected()){
				exp.shuffleList = new RandomList<Card>();
				for (Card card : exp.cards){
					if (!card.selected || card.disabled) continue;
					exp.shuffleList.add(card);
				}
				exp.minVal = exp.page.getMin();
				exp.maxVal = Math.min(exp.page.getMax(), exp.shuffleList.size());
				if (exp == CardData.ALCHEMY && alchemy){	// 錬金術の抽出不可能性
					if (exp.minVal > 0) exp.minVal = 3;
					if (exp.maxVal < 3) exp.maxVal = 0;
				}
				if (exp.minVal > exp.maxVal) throw new ExpectedException(exp.page.title + "セット: 抽出できる枚数が、抽出すべき最低枚数に足りません");
				minSum += exp.minVal;
				maxSum += exp.maxVal;
			} else {
				exp.minVal = exp.maxVal = 0;
			}
		}
		if (minSum > needed) throw new ExpectedException("各セットから抽出すべき最低枚数の合計が、「1回のシャッフルで抽出する枚数」を超えています");
		if (maxSum < needed) throw new ExpectedException("各セットから抽出できる枚数の合計が、「1回のシャッフルで抽出する枚数」に足りません");

		/* TRASH, ACTION, BUY の条件による抽出 */
		try {
			chosen = constraints(needed - minSum, trash, action, buy);	// Card#disabled, Expansion#minVal, Expansion#maxVal が更新される
		} catch (Exception ex){
			throw new ExpectedException("抽出可能なカードの中に条件を満たすカードの組合せは存在しません");
		}
		needed -= chosen.size();

		/* それ以外の条件による抽出 */
		try {
			/* 各セット毎の最低枚数の制約を満たす */
			for (Expansion exp : Expansion.allInstances){
				if (!exp.isSelected()) continue;
				while (exp.minVal > 0){
					while ((tmp = exp.shuffleList.randomRemove()).disabled);
					chosen.add(tmp); tmp.disabled = true; exp.maxVal--; exp.minVal--; needed--;
				}
			}

			/* 残りの抽出 */
			RandomList<Card> allList = new RandomList<Card>();
			for (Expansion exp : Expansion.allInstances){
				if (!exp.isSelected() || exp.maxVal <= 0) continue;
				for (Card card : exp.shuffleList) if (card.isAvailable()) allList.add(card);
			}
			if (alchemy){	// 錬金術を入れるなら3枚以上
				while (needed > 2){
					while (!(tmp = allList.randomRemove()).isAvailable());
					chosen.add(tmp); tmp.disabled = true; tmp.expansion.maxVal--; needed--;
				}
				int alchemyCount = 0;
				for (Card card : chosen) if (card.expansion == CardData.ALCHEMY) alchemyCount++;
				if (alchemyCount == 0){	// 残り2枚の時点で錬金術ナシ : 残りは錬金術以外から選ぶ
					while (needed > 0){
						while (!(tmp = allList.randomRemove()).isAvailable() || tmp.expansion == CardData.ALCHEMY);
						chosen.add(tmp); tmp.disabled = true; tmp.expansion.maxVal--; needed--;
					}
				} else {	// 残り2枚の時点で錬金術アリ : 錬金術が3枚に達するまでは錬金術から選ぶ
					while (alchemyCount < 3){
						while (!(tmp = CardData.ALCHEMY.shuffleList.randomRemove()).isAvailable());
						chosen.add(tmp); tmp.disabled = true; tmp.expansion.maxVal--; needed--; alchemyCount++;
					}
				}
			}
			while (needed > 0){
				while (!(tmp = allList.randomRemove()).isAvailable());
				chosen.add(tmp); tmp.disabled = true; tmp.expansion.maxVal--; needed--;
			}
		} catch (Exception ex){
			ex.printStackTrace();
			throw new ExpectedException("予期せぬエラーが発生しました");	// 発生しない筈
		}

		return chosen;
	}

	private static TriState getColonyState(List<Card> chosen){
		if (!CardData.PROSPERITY.isSelected()) return TriState.NO;

		int prosperity = 0;
		for (Card card : chosen) if (card.expansion == CardData.PROSPERITY) prosperity++;
		if (RandomList.mt.genrand_real2() * chosen.size() < prosperity) return TriState.TRUE;
		return TriState.FALSE;
	}

	private static TriState getShelterState(List<Card> chosen){
		if (!CardData.DARKAGES.isSelected()) return TriState.NO;

		int darkages = 0;
		for (Card card : chosen) if (card.expansion == CardData.DARKAGES) darkages++;
		if (RandomList.mt.genrand_real2() * chosen.size() < darkages) return TriState.TRUE;
		return TriState.FALSE;
	}

	public String toString(){
		String ret = "";
		int i = 0;
		for (Card card : chosen){
			if (ret != "") ret += ", ";
			if (++i > 5) return ret + "...";
			ret += card.name;
		}
		return ret;
	}

	public List<Card> chosen;
	private List<Card> chosen2;
	public Card bane = null;
	public boolean bmarket = false;
	public TriState useColonyAndPlatinum = TriState.FALSE, useShelters = TriState.FALSE;
	public String alert = null;
	public Shuffler(int needed, boolean trash, boolean action, boolean buy, boolean bmRange, boolean alchemy) throws ExpectedException {
		/********** 第1シャッフル **********/
		/* 災いカードとして抽出され得るものが無いならば予め魔女娘は無効化しておく */
		RandomList<Card> possiblyBane = new RandomList<Card>();
		for (Expansion exp : Expansion.allInstances){
			if (exp.isSelected()){
				for (Card card : exp.cards){
					card.disabled = false;
					if (card.selected && (card.attrib & BANE) != 0) possiblyBane.add(card);
				}
			} else {
				for (Card card : exp.cards) card.disabled = false;
			}
		}
		if (possiblyBane.isEmpty()) CardData.YWITCH.disabled = true;

		/* シャッフル */
		chosen = shuffle(needed, trash, action, buy, alchemy);	// 失敗した場合は例外発生

		/* 災いカード・闇市場デッキの必要性を判定 */
		boolean ywitchInSupply = false, bmarketInSupply = false;
		for (Card card : chosen){
			if (card == CardData.YWITCH){
				ywitchInSupply = true;
			} else if (card == CardData.BMARKET){
				bmarketInSupply = true;
			}
		}
		boolean ywitchInBMarket = bmarketInSupply && !possiblyBane.isEmpty() && CardData.YWITCH.expansion.isSelected() && (CardData.YWITCH.selected || !bmRange);
		if (!(ywitchInSupply || ywitchInBMarket)){	// 災いカードは不要
			bmarket = bmarketInSupply;
			useColonyAndPlatinum = getColonyState(chosen);
			useShelters = getShelterState(chosen);
			return;
		}

		/* 災いカードの抽出 */
		try {
			while (!(bane = possiblyBane.randomRemove()).isAvailable());	// ここで失敗した場合は第2シャッフルへ
			bane.disabled = true;
			bmarket = bmarketInSupply || bane == CardData.BMARKET;
			useColonyAndPlatinum = getColonyState(chosen);
			useShelters = getShelterState(chosen);
			return;
		} catch (Exception ex){}
		chosen2 = chosen;


		/********** 第2シャッフル（災いカードを予め定めた上で再度シャッフル） **********/
		int needed2 = needed;
		possiblyBane.clear();
		bane = null;

		for (Expansion exp : Expansion.allInstances){
			if (exp.isSelected()){
				for (Card card : exp.cards){
					card.disabled = false;
					if (card.selected && (card.attrib & BANE) != 0) possiblyBane.add(card);
				}
			} else {
				for (Card card : exp.cards) card.disabled = false;
			}
		}
		if (ywitchInSupply){	// 1回目のシャッフルで魔女娘が抽出されていれば魔女娘をサプライに含める
			CardData.YWITCH.disabled = true;
			needed2--;
		}
		if (bmarketInSupply){	// 1回目のシャッフルで闇市場が抽出されていれば闇市場をサプライに含める
			possiblyBane.remove(CardData.BMARKET);
			if (possiblyBane.isEmpty()){	// 但し災いカードとして抽出され得るカードが闇市場以外に無い場合は闇市場を災いカードにする
				bmarketInSupply = false;
				bane = CardData.BMARKET;
			} else {
				CardData.BMARKET.disabled = true;
				needed2--;
			}
		}
		try {
			if (bane == null) bane = possiblyBane.randomRemove();	// 災いカードの抽出
			bane.disabled = true;
			if ((bane.attrib & TRASH) != 0) trash = false;
			if ((bane.attrib & ACTION) != 0) action = false;
			if ((bane.attrib & BUY) != 0) buy = false;
		} catch (Exception ex){
			ex.printStackTrace();
			throw new ExpectedException("予期せぬエラーが発生しました");	// 発生しない筈
		}

		/* シャッフル */
		try {
			chosen = shuffle(needed2, trash, action, buy, alchemy);	// ここで失敗した場合は第3シャッフルへ
			if (ywitchInSupply) chosen.add(CardData.YWITCH);
			if (bmarketInSupply) chosen.add(CardData.BMARKET);
			bmarket = bmarketInSupply || bane == CardData.BMARKET;
			useColonyAndPlatinum = getColonyState(chosen);
			useShelters = getShelterState(chosen);
			return;
		} catch (Exception ex){}


		/********** 第3シャッフル（災いカードの抽出は不可能なので魔女娘を無効化して再度シャッフル） **********/
		for (Expansion exp : Expansion.allInstances){
			for (Card card : exp.cards) card.disabled = false;
		}
		CardData.YWITCH.disabled = true;

		/* シャッフル */
		try {
			chosen = shuffle(needed, trash, action, buy, alchemy);
			bane = null;
			bmarket = false;
			useColonyAndPlatinum = getColonyState(chosen);
			useShelters = getShelterState(chosen);
			return;
		} catch (Exception ex){
			chosen = chosen2;	// 魔女娘は必要だが災いカードは抽出できない
			bane = null;
			bmarket = false;
			useColonyAndPlatinum = getColonyState(chosen);
			useShelters = getShelterState(chosen);
			alert = ywitchInSupply ? "サプライに魔女娘が含まれますが、災いカードが抽出できません" : "闇市場デッキに魔女娘が含まれますが、災いカードが抽出できません";
			return;
		}
	}
}

